#include <nag_rwrapper_utils.h>

int nag_extractDoubleVector(int n, SEXP v, double cv[]) {
  /* nag_extractDoubleVector: Take an R object and extract a double precision vector.

     On entry:
     n   - Length of the vector
     v   - R object holding the vector
     
     On exit:
     cv  - The vector

     Returns:
     error code, 0 = All OK, anything else implies an error in input
     vector, 1 = Not a vector, 2 = Not numeric, 3 = Length != n,            
     4 = An entry was NaN or NA
  */
  int failed, i;
  double *dv;
  SEXP this_dv;

  if (!(failed = nag_isNumericVector(n,v))) {
    /* Cast the input vector into a vector of doubles */
    PROTECT(this_dv = coerceVector(v, REALSXP));

    /* Loop over each entry in v */
    dv = REAL(this_dv);
    for (i = 0; i < n; i++) {
      /* Check the entry is a valid double */
      if (!R_FINITE(dv[i])) {
        failed = 4;
        break;
      }

      /* Copy the entry into the output array */
      cv[i] = dv[i];
    }
    UNPROTECT(1);
  }

  return failed;
}

int nag_extractIntegerVector(int n, SEXP v, Integer cv[]) {
  /* nag_extractIntegerVector: Take an R object and extract an Integer vector.

     On entry:
     n   - Length of the vector
     v   - R object holding the vector
     
     On exit:
     cv  - The vector

     Returns:
     error code, 0 = All OK, anything else implies an error in input
     vector, 1 = Not a vector, 2 = Not numeric, 3 = Length != n,            
     4 = An entry was NaN or NA
  */
  int failed, i;
  int *iv;
  SEXP this_iv;

  if (!(failed = nag_isNumericVector(n,v))) {
    /* Cast the input vector into a vector of int's */
    PROTECT(this_iv = coerceVector(v,INTSXP)); 

    /* Loop over each entry in v */
    iv = INTEGER(this_iv);
    for (i = 0; i < n; i++) {
      /* Check the entry is a valid Integer */
      if (iv[i] == NA_INTEGER) {
        failed = 4;
        break;
      }
      /* Copy the entry into the output array */
      cv[i] = (int) iv[i];
    }
    UNPROTECT(1);
  }
  return failed;
}

int nag_isNumericVector(int n, SEXP v) {
  /* nag_isNumericVector: Checks that an R object is a numeric vector

     On entry:
     n   - Length of the vector
     v   - R object holding the vector

     Returns
     error code, 0 = All OK, anything else implies an error in input
     vector, 1 = Not a vector, 2 = Not numeric, 3 = Length != n
  */        
  int failed = 0;  

  if (!isVector(v)) {
    failed = 1;
  } else if (!isNumeric(v)) {
    failed = 2;
  } else if (LENGTH(v) != n) {
    failed = 3;
  } else {
    failed = 0;
  }
    
  return failed;
}

int nag_extractDoubleMatrix(int cmajor, int nrow, int ncol, SEXP m, double *cm) {
  /* nag_extractDoubleMatrix: Take an R object and extract a double precision matrix.

     On entry:
     cmajor - If set to 0 then the extracted matrix is in row major order,
              else it is in column major order
     nrow   - Number of rows
     ncol   - Number of columns
     m      - R object holding the matrix
     
     On exit:
     cm  - The matrix

     Returns:
     error code, 0 = All OK, anything else implies an error in input
     vector, 1 = Not a matrix, 2 = Not numeric, 3 = Number of rows != nrow,
     4 = Number of columns != ncol, 5 = An entry was NaN or NA
  */
  int failed, i, j, k, pk;
  double *dm;

  if (!(failed = nag_isNumericMatrix(nrow,ncol,m))) {
    dm = REAL(m);
    if (cmajor) {

      /* Column major order (think this is the order R stores things in ???) */
      for (j = k = 0; j < ncol; j++) {               
        for (i = 0; i < nrow; i++, k++) {
          if (!R_FINITE(dm[k])) {
            failed = 5;
            break;
          }
          cm[k] = dm[k];
        }
      }

    } else {
      /* Row major order */
      for (j = k = 0; j < ncol; j++) {        
        for (i = 0; i < nrow; i++, k++) {        
          if (!R_FINITE(dm[k])) {
            failed = 5;
            break;
          }
          pk = i * ncol + j;
          cm[pk] = dm[k];
        }
      }      
    }
  }

  return failed;
}

int nag_isNumericMatrix(int nrow, int ncol, SEXP m) {
  /* nag_isNumericMatrix: Take an R object and extract a double precision matrix.

     On entry:
     nrow   - Number of rows
     ncol   - Number of columns
     
     Returns:
     error code, 0 = All OK, anything else implies an error in input
     vector, 1 = Not a matrix, 2 = Not numeric, 3 = Number of rows != nrow,
     4 = Number of columns != ncol
  */
  int failed = 0;

  if (!isMatrix(m)) {
    failed = 1;
  } else if (!isNumeric(m)) {
    failed = 2;
  } else if (nrows(m) != nrow) {
    failed = 3;
  } else if (ncols(m) != ncol) {
    failed = 4;
  } else {
    failed = 0;
  }

  return failed;
}
    
SEXP nag_getListElement(SEXP list, const char *str) {
  /* nag_getListElement: Returns the list element named str, or returns NULL.

     This is based on the function getListElement in "Writing R Extensions"
  */

  SEXP elmt, names;
  int i;

  elmt = R_NilValue;
  names = getAttrib(list, R_NamesSymbol);

  for (i = 0; i < length(list); i++) {
    if (!strcmp(CHAR(STRING_ELT(names,i)),str)) {
      elmt = VECTOR_ELT(list,i);
      break;
    }
  }
  return elmt;
}


SEXP nag_md2r(int cmajor, int nrow, int ncol, double m[]) {
  /* nag_md2r: Take a matrix of doubles and return an matrix R object
     cmajor - If 0 then matrix m is in row major order, else m is in
              column major order
     nrow   - Number of rows
     ncol   - Number of columns
     m      - Array of doubles
  */
  SEXP ans;
  double *dans;
  int i, j, k, pk;

  /* Protect the vector ans from R's garbage collector */
  PROTECT(ans = allocMatrix(REALSXP, nrow, ncol));
  dans = (double *) REAL(ans);
  if (cmajor) {
    for (j = k = 0; j < ncol; j++) {
      for (i = 0; i < nrow; i++, k++) {
        dans[k] = m[k];
      }
    }
  } else {
    for (j = k = 0; j < ncol; j++) {
      for (i = 0; i < nrow; i++, k++) {
        pk = i * ncol + j;
        dans[k] = m[pk];
      }
    }
  }

  /* Unprotect the last 1 object protected (which in this is ans) */
  UNPROTECT(1);

  return ans;
}


SEXP nag_i2r(int n, Integer x[]) {
  /* nag_i2r: Take an array of Integers, and return an INTEGER R object
     n  - Number of elements
     x  - Array of doubles
  */
  SEXP ans;
  int *ians, i;

  /* Protect the vector ans from R's garbage collector */
  PROTECT(ans = allocVector(INTSXP, n));
  ians = INTEGER(ans);
  for (i = 0; i < n; i++) {
    ians[i] = (int) x[i];
  }

  /* Unprotect the last 1 object protected (which in this is ans) */
  UNPROTECT(1);

  return ans;
}

SEXP nag_fi2r(int n, NAG_fortran_integer x[]) {
  /* nag_fi2r: Take an array of NAG_fortran_integer, and return an INTEGER R object
     n  - Number of elements
     x  - Array of doubles
  */
  SEXP ans;
  int *ians, i;

  /* Protect the vector ans from R's garbage collector */
  PROTECT(ans = allocVector(INTSXP, n));
  ians = INTEGER(ans);
  for (i = 0; i < n; i++) {
    ians[i] = (int) x[i];
  }

  /* Unprotect the last 1 object protected (which in this is ans) */
  UNPROTECT(1);

  return ans;
}

SEXP nag_l2l(int n, Nag_Boolean x[]) {
  /* nag_l2l: Take an array of Nag_Boolean, and return a LGLSXP R object
     n  - Number of elements
     x  - Array of doubles
  */
  SEXP ans;
  int *ians;
  int i;

  /* Protect the vector ans from R's garbage collector */
  PROTECT(ans = allocVector(LGLSXP, n));
  for (i = 0; i < n; i++) {
    ians = LOGICAL(ans);
    if (x[i] == Nag_TRUE) {
      ians[i] = 1;
    } else {
      ians[i] = 0;
    }
  }

  /* Unprotect the last 1 object protected (which in this is ans) */
  UNPROTECT(1);

  return ans;
}


SEXP nag_d2r(int n, double x[]) {
  /* nag_d2r: Take an array of doubles, and return a REAL R object
     n  - Number of elements
     x  - Array of doubles
  */
  SEXP ans;
  double *rans;
  int i;

  /* Protect the vector ans from R's garbage collector */
  PROTECT(ans = allocVector(REALSXP, n));
  rans = REAL(ans);
  for (i = 0; i < n; i++) {
    rans[i] = x[i];
  }

  /* Unprotect the last 1 object protected (which in this is ans) */
  UNPROTECT(1);

  return ans;
}

SEXP nag_2d2r(int n1, double x1[], int n2, double x2[]) {
  /* nag_d2r: Take an two arrays of doubles, and return a REAL R object
     n1 - Number of elements in first array
     x1 - First array of doubles
     n2 - Number of elements in second array
     x2 - Second array of doubles

     Returns and REAL vector equivalent to c(x1,x2)
  */
  SEXP ans;
  double *rans;
  int i, j;

  /* Protect the vector ans from R's garbage collector */
  PROTECT(ans = allocVector(REALSXP, n1 + n2));
  rans = REAL(ans);
  for (j = 0; j < n1; j++) {
    rans[j] = x1[j];
  }
  for (i = 0; i < n2; i++, j++) {
    rans[j] = x2[i];
  }

  /* Unprotect the last 1 object protected (which in this is ans) */
  UNPROTECT(1);

  return ans;
}

/*************************************************************************************/
/* Routines for storing and retrieving a pointer in an array of integerrs */
#ifdef WIN64
#define MY_NAG_ULONG_TYPE unsigned __int64
#else
#define MY_NAG_ULONG_TYPE unsigned long
#endif

void nag_convert_p2i(SEXP p, NAG_fortran_integer *v) {
  /* NAG_CONVERT_P2I: Convert an SEXP object into two integers.
     On entry:
     p - Pointer to SEXP object

     On exit:
     v - Array to hold the converted pointer, v[2]
  */
  MY_NAG_ULONG_TYPE lp;

  lp = (MY_NAG_ULONG_TYPE) p;

#ifdef IS_64BIT
  v[0] = (NAG_fortran_integer) (lp >> 32);
  v[1] = (NAG_fortran_integer) (lp & 0xffffffff);
#else
  /* Handle 32-bit case separately just to avoid gcc compiler
     warning about shifting a 32-bit quantity by 32 bits */
  v[0] = 0;
  v[1] = (NAG_fortran_integer) lp;
#endif
}

void nag_convert_i2p(NAG_fortran_integer *v, SEXP *p) {
  /* NAG_CONVERT_I2P: Convert two integers into an SEXP object
     On entry:
     v - Array to holding the converted pointer, v[2]

     On exit:
     p - Pointer to function
  */
  MY_NAG_ULONG_TYPE lp;

  if (v[0] != 0 || v[1] != 0)
    {
#ifdef IS_64BIT
      /* N.B. weird casting of v[1] to acml_uint_t is because without it,
         if the top bit of v[1] happens to be set, a cast directly to unsigned
         long will propagate the 1 bit all the way to the 64th bit */
      lp = ((MY_NAG_ULONG_TYPE) v[0] << 32) | ((MY_NAG_ULONG_TYPE) ((NAG_fortran_integer)v[1]));
      *p = (SEXP) lp;
#else
      /* Handle 32-bit case separately just to avoid gcc compiler
         warning about shifting a 32-bit quantity by 32 bits */
      lp = (MY_NAG_ULONG_TYPE) v[1];
      *p = (SEXP) lp;
#endif
    }
  else
    *p = NULL;
}

#undef MY_NAG_ULONG_TYPE
/* Routines for storing and retrieving a pointer in an array of integerrs */
/*************************************************************************************/

/*************************************************************************************/
/* The following functions are extensions of the lang2 to lang4 and list1 to list 4
   in the Rinlinedfuns.h. 

   These are only needed when use callback functions that are the function themselves
   (rather than the body of the function). Probably better to just stick with using
   the body of the function!
*/
SEXP lang5(SEXP s, SEXP p1, SEXP p2, SEXP p3, SEXP p4)
{
    PROTECT(s);
    s = LCONS(s, list4(p1, p2, p3, p4));
    UNPROTECT(1);
    return s;
}

SEXP lang6(SEXP s, SEXP p1, SEXP p2, SEXP p3, SEXP p4, SEXP p5)
{
    PROTECT(s);
    s = LCONS(s, list5(p1, p2, p3, p4, p5));
    UNPROTECT(1);
    return s;
}

SEXP lang7(SEXP s, SEXP p1, SEXP p2, SEXP p3, SEXP p4, SEXP p5, SEXP p6)
{
    PROTECT(s);
    s = LCONS(s, list6(p1, p2, p3, p4, p5, p6));
    UNPROTECT(1);
    return s;
}

SEXP lang8(SEXP s, SEXP p1, SEXP p2, SEXP p3, SEXP p4, SEXP p5, SEXP p6, SEXP p7)
{
    PROTECT(s);
    s = LCONS(s, list7(p1, p2, p3, p4, p5, p6, p7));
    UNPROTECT(1);
    return s;
}

SEXP list5(SEXP s, SEXP l1, SEXP l2, SEXP l3, SEXP l4)
{
    PROTECT(s);
    s = CONS(s, list4(l1, l2, l3, l4));
    UNPROTECT(1);
    return s;
}

SEXP list6(SEXP s, SEXP l1, SEXP l2, SEXP l3, SEXP l4, SEXP l5)
{
    PROTECT(s);
    s = CONS(s, list5(l1, l2, l3, l4, l5));
    UNPROTECT(1);
    return s;
}

SEXP list7(SEXP s, SEXP l1, SEXP l2, SEXP l3, SEXP l4, SEXP l5, SEXP l6)
{
    PROTECT(s);
    s = CONS(s, list6(l1, l2, l3, l4, l5, l6));
    UNPROTECT(1);
    return s;
}
/*************************************************************************************/


