#include <R.h>
#include <Rinternals.h>

#include "nag_fortran_routines.h"
#include "nag_rwrapper_utils.h"

#ifdef __cplusplus
extern "C" {
#endif
  extern void __stdcall we04abfF(double *xc, double *fc, NAG_fortran_integer *iuser, double *ruser);
  extern void __stdcall we04abfL(double *xc, double *fc, NAG_fortran_integer *iuser, double *ruser);
#ifdef __cplusplus
}
#endif

SEXP we04abf(SEXP funct, SEXP ab,  SEXP e, SEXP mfun, SEXP rho) {
  /* we04abf: Wrapper for the NAG C library routine e04abf.
     On entry:
     funct  - Either an R function or the body of a function which we wish to minimise
              The R function pointed to by comm->p[0] can either be a function, or
              the body of a function, i.e. if
        
              ff = function(y) {sin(y) / y}

              then can either pass ff directly, or could pass
                     {as.double(ff(x))}
              or
               gg = function(x) {
                  as.double(ff(x))
               }
              and then use body(gg)

              In the case of the code segment, or using the body function, the variable x 
              must be used to denote the value at which the function is to be evaluated.

              Normally would only use one of the above, and hence could remove the unused
              code from this function

     ab     - Vector of two elements containing the interval holding the minimum, 
              this is equivalent to c(A,B), where "c" is the R function of same name
              and "A" and "B" are arguments to e04abf.
     e      - Vector of two elements containing the relative and absolute errors. This is
              equivalent to c(E1,E2), where "E1" and "E2" are arguments to e04abf. If
              e = c(NA,NA) the default values are used for E1 and E2, similarly if e only
              has one value then a default is used for E2.
     mfun   - Integer holding the maximum number of function evaluations (equivalent to
              c(MAXCAL), where "MAXCAL" is an argument to e04abf. If mfun = c(NA), then
              a default value of 100 is used.
     rho    - R environment used to evaluate the function

     Returns:
     An R list, with elements "x", "f" and "ab", where "x" and "f" are equivalent to the
     output arguements of e04abf of the same name (i.e. the estimated minimum and the 
     value of the function at x), and "ab" is the equivalent to c(a,b) on exit from e04abf, i.e.
     the interval containing the estimated minimum.
  */
  double e1, e2, a, b, t, x, f;
  NAG_fortran_integer maxcal, ifail;
  NAG_fortran_integer iuser[5];
  double ruser[1];
  SEXP ans, names;
  char line[100];

  /* Perform some initial checks */
  if (LENGTH(ab) != 2) {
    error("ab must be a numeric vector of length 2");
  }
  if (!isLanguage(funct) && !isFunction(funct)) {
    error("funct must be a function or the body of a function");
  }
  if (!isEnvironment(rho)) {
    error("rho must be an environment");
  }
  /* End of parameter checks */

  /* Get initial bounds for the solution */
  a = REAL(ab)[0];
  b = REAL(ab)[1];
  if (a > b) {
    t = a;
    a = b;
    b = t;
  }
  if (!R_FINITE(a) || !R_FINITE(b)) {
    error("ab must be a numeric vector of length 2");
  }

  /* Get required accuracy, or use defaults if not supplied */  
  e1 = REAL(e)[0];
  if (!R_FINITE(e1)) e1 = 0.0;
  if (LENGTH(e) > 1) {
    e2 = REAL(e)[1];
    if (!R_FINITE(e2)) e2 = 0.0;
  } else {
    e2 = 0.0;
  }
  
  /* Get the maximum number of iterations, or use default of 100 if not supplied */
  maxcal = INTEGER(mfun)[0];
  if (maxcal == NA_INTEGER) maxcal = 100;

  /* Convert function pointer funct and rho into a couple of integers and store in IUSER */
  nag_convert_p2i(funct, &iuser[0]);
  nag_convert_p2i(rho, &iuser[2]);

  /* Error code used if the callback function does not return a numeric value */
  iuser[4] = 0;

  /* Call the NAG routine */
  ifail = 1;
  if (isLanguage(funct)) {
    /* Run version that expects a body of a function */
    E04ABA(we04abfL, &e1, &e2, &a, &b, &maxcal, &x, &f, iuser, ruser, &ifail);
  } else if (isFunction(funct)) {
    /* Run version that expects the function itself */
    E04ABA(we04abfF, &e1, &e2, &a, &b, &maxcal, &x, &f, iuser, ruser, &ifail);
  }

  /* Check the error returns */
  if (iuser[4] != 0) {
    error("At least one call to the function funct returned a non-numeric value");
  }
  /* Error returns from NAG routine */
  if (ifail != 0) {
    sprintf(line,"Error in NAG library routine E04ABA, IFAIL on exit = %ld\n", (long) ifail);
    error(line);
  }

  /* Get the list of results */
  PROTECT(ans = allocVector(VECSXP,3));
  /* Return the estimated point of the minimum (x) */
  SET_VECTOR_ELT(ans,0,nag_d2r(1,&x));
  /* Return the value of F at x (f) */
  SET_VECTOR_ELT(ans,1,nag_d2r(1,&f));
  /* Return the lower and upper bounds for x */
  SET_VECTOR_ELT(ans,2,nag_2d2r(1,&a,1,&b));

  /* Get names for the list of results */
  PROTECT(names = allocVector(STRSXP, 3));
  SET_STRING_ELT(names,0,mkChar("x"));
  SET_STRING_ELT(names,1,mkChar("f"));
  SET_STRING_ELT(names,2,mkChar("ab"));
  setAttrib(ans,R_NamesSymbol,names);

  /* Unprotect the ans and names objects */
  UNPROTECT(2);

  return(ans);
}
  

void __stdcall we04abfF(double *xc, double *fc, NAG_fortran_integer *iuser, double *ruser) {
  /* we04abfF: Wrapper for the callback function to E04ABF. For use
     when the user has supplied a function (as opposed to the body of a function)
     Used when calling a NAG Fortran routine
     
     On entry:
     xc    - the point at which the function is to be evaluated
     iuser - Integer communication array
             iuser[0:1] - Holds integer representation of pointer to the function (or 
                          code segment) that is being evaluated.
             iuser[2:3] - Holds integer representation of pointer to the environment
                          the function is being evaluated in
     ruser - Real communication array, not used

     On exit:
     fc    - the value of the function at xc
     iuser - Integer communication array
             iuser[4]   - set to 1 if the function returns a non-numeric value
                          In these cases, fc is set to zero.
     
  */
  SEXP funct, rho, R_fcall;

  /* Get the pointer to the R function */
  nag_convert_i2p(&iuser[0], &funct);
  /* Get the pointer to the R environment */
  nag_convert_i2p(&iuser[2], &rho);

  /* In this case funct points to an actual function that takes a single argument */
  PROTECT(R_fcall = lang2(funct,R_NilValue));
  SETCADR(R_fcall, nag_d2r(1,xc));
  
  /* Get the results from evaluating the function in R */
  *fc = REAL(eval(R_fcall, rho))[0];
  
  /* Release R_fcall */
  UNPROTECT(1);

  if (!R_FINITE(*fc)) {
    /* Returned value is not valid. Can't exit NAG routine cleanly, so 
       set the 5th element of IUSER, and set returned value to 0
    */
    iuser[4] = 1;
    *fc = 0.0;
  }
}

void __stdcall we04abfL(double *xc, double *fc, NAG_fortran_integer *iuser, double *ruser) {
  /* we04abfL: Wrapper for callback function to E04ABF. For use
     when the user has supplied the body of a function (as opposed to the whole function)
     Used when calling a NAG Fortran routine
     
     On entry:
     xc    - the point at which the function is to be evaluated
     iuser - Integer communication array
             iuser[0:1] - Holds integer representation of point to the function (or code
                          code segment) that is being evaluated.
             iuser[2:3] - Holds integer representation of point to the environment
                          the function is being evaluated in
     ruser - Real communication array, not used

     On exit:
     fc    - the value of the function at xc
     iuser - Integer communication array
             iuser[4]   - set to 1 if the function returns a non-numeric value
                          In these cases, fc is set to zero.
     
  */
  SEXP funct, rho;

  /* Get the pointer to the R function */
  nag_convert_i2p(&iuser[0], &funct);
  /* Get the pointer to the R environment */
  nag_convert_i2p(&iuser[2], &rho);

  /* Makes xc available to the function as a variable named x */
  defineVar(install("x"), nag_d2r(1,xc), rho);

  /* Get the results from evaluating the function in R */
  *fc = REAL(eval(funct,rho))[0];

  if (!R_FINITE(*fc)) {
    /* Returned value is not valid. Can't exit NAG routine cleanly, so 
       set the 5th element of IUSER, and set returned value to 0
    */
    iuser[4] = 1;
    *fc = 0.0;
  }
}
