#include <R.h>
#include <Rinternals.h>

#include "nag_fortran_routines.h"
#include "nag_rwrapper_utils.h"

/* Wrappers for the user supplied functions for the constraints and objective functions */
#ifdef __cplusplus
extern "C" {
#endif
  void __stdcall we04ucfL_con(NAG_fortran_integer *mode, NAG_fortran_integer *ncnln, NAG_fortran_integer *n, NAG_fortran_integer *ldcj, 
                              NAG_fortran_integer needc[], double x[], double c[], double cjac[], NAG_fortran_integer *nstate, 
                              NAG_fortran_integer iuser[], double ruser[]);
  void __stdcall we04ucfL_obj(NAG_fortran_integer *mode, NAG_fortran_integer *n, double x[], double *objf, double objgrd[], NAG_fortran_integer *nstate, 
                              NAG_fortran_integer iuser[], double ruser[]);
  void __stdcall we04ucfF_con(NAG_fortran_integer *mode, NAG_fortran_integer *ncnln, NAG_fortran_integer *n, NAG_fortran_integer *ldcj,
                              NAG_fortran_integer needc[], double x[], double c[], double cjac[], NAG_fortran_integer *nstate,
                              NAG_fortran_integer iuser[], double ruser[]);
  void __stdcall we04ucfF_obj(NAG_fortran_integer *mode, NAG_fortran_integer *n, double x[], double *objf, double objgrd[], NAG_fortran_integer *nstate,
                              NAG_fortran_integer iuser[], double ruser[]);
#ifdef __cplusplus
}
#endif

/* Utility functions */
void we04ucf_clean_up(double *pd1, double *pd2, double *pd3, double *pd4, double *pd5, double *pd6, double *pd7,
                      double *pd8, double *pd9, double *pd10, NAG_fortran_integer *pi1, NAG_fortran_integer *pi2);
int we04ucf_handle_options(SEXP optlist, NAG_fortran_integer lwsav[], NAG_fortran_integer iwsav[], double rwsav[]);

/* Length of array to hold error messages, no checks are made to ensure this is large enough */
#define LEN_ERR_MSG 100
/* Flag indicating column major order (!=0) or row major order (0) */
#define IS_COLUMN_MAJOR_ORDER 1

SEXP we04ucf(SEXP psize, SEXP a, SEXP lower, SEXP upper, SEXP objfun, SEXP confun, SEXP initial, SEXP optlist, SEXP rho) {
  /* we04ucf: Wrapper for NAG Fortran library routine e04ucf.

     On entry:
     psize   - Vector containing the problem size, psize = c(N,NCLIN,NCNLN),
               where N = number of variables, NCLIN = number linear constraints
               NCNLN = number of non-linear constraints
     a       - (NCLIN x N) Matrix of linear constraints. If NCLIN = 0, a is not referenced, 
     lower   - List of lower constraint limits. List should contain three elements, the first
               a vector of N limits for simple constraints, the second a vector of NCLIN
               limits for the linear constraints and the third a vector of NCNLN limits for
               the non-linear constraints
     upper   - List of upper constraint limits. This should be in the same form as lower
     objfun  - Either an R function or the body of an R function that calculates the objective 
               function and its gradients.
     confun  - Either an R function or the body of an R function that calculates the non-linear 
               constraints and corresponding derivatives.
     initial - Vector of N initial values
     optlist - List of options, in format that can be read by we04ucf_handle_options
     rho     - R environment in which the two functions objfun and confun are to be evaluated.


     Returns:
     A list with entries: 'objf', 'x' and 'g', which correspond to the e04ucf arguments of the
     same name. If, on entry, optlist is a list (irrespective of whether it has any valid options), 
     then 'h', 'iter', 'nf', 'conf' and 'conjac' are also returned. 'conf' and 'conjac' are only
     returned if NCNLN > 0. These last 5 parameters correspond to the optional arguments of the same
     name held in the e04ucf argument options.

     Notes:
     Much of the parameter checking done in this routine could have been done in the R wrapper
     that calls we04ucf (and would probably have been simpler to do). 
  */
  NAG_fortran_integer nclin, n, ncnln, lans, cmajor, failed, nbounds, k, ifail, ldcj, lda, size_cjac, 
    size_c, liwork, lwork, size_ca, ldr, iter, nadv, iflag;
  NAG_fortran_integer *istate = 0, *iwork = 0;
  Integer probsize[3];
  SEXP ans, names, lower_s, lower_l, lower_n, upper_s, upper_l, upper_n; 
  double objf;
  double *ca = 0, *bu = 0, *bl = 0, *x = 0, *objgrd = 0, *r = 0, *cjac = 0, *c = 0, *work = 0, *clamda = 0;
  char err_msg[LEN_ERR_MSG];
  double ruser[1];
  NAG_fortran_integer iuser[6];

  NAG_fortran_integer llwsav = 120, liwsav = 610, lrwsav = 475, lcwsav = 1;
  NAG_fortran_integer iwsav[610], lwsav[120];
  double rwsav[475];
  char rname[6] = "E04UCA";
  char cwsav[80];
  int unit_num;

  /* Set type of matrix we are working with (when calling NAG C library this will be row major) */
  cmajor = IS_COLUMN_MAJOR_ORDER;

  /* Initial parameter checks */
  if ((failed = nag_extractIntegerVector(3,psize,&probsize[0]))) {
    sprintf(err_msg,"'psize' must be a numeric vector of length 3 (code = %ld)", (long) failed);
    error(err_msg);
  }

  /* NB: Need to use isNewList in preference to isList to test for lists */
  if (!isNewList(lower)) {
    error("'lower' must be a list with three elements");
  }
  if (length(lower) != 3) {
    error("'lower' must be a list with three elements");
  }
  if (!isNewList(upper)) {
    error("'upper' must be a list with three elements");
  }
  if (length(upper) != 3) {
    error("'upper' must be a list with three elements");
  }
  if (!isLanguage(confun) && !isFunction(confun)) {
    error("'confun' must be a function or the body of a function");
  }
  if (!isLanguage(objfun) && !isFunction(objfun)) {
    error("'objfun' must be a function or the body of a function");
  }
  if (isLanguage(objfun) != isLanguage(confun)) {
    error("'objfun' and 'confun' should either both be functions, or both the body of a function");
  }
  if (!isEnvironment(rho)) {
    error("'rho' must be an environment");
  }
  /* End of initial parameter checks */

  /* Get the problem size */
  n = probsize[0];
  nclin = probsize[1];
  ncnln = probsize[2];
  if (n < 0 || nclin < 0 || ncnln < 0) {
    error("'psize' must be a numeric vector of positive values");
  }

  /* Allocate memory for the matrix of linear constraints */
  size_ca = nclin*n;
  if (size_ca < 1) size_ca = 1;  
  if (!(ca = (double *) malloc(size_ca*sizeof(double)))) {
     we04ucf_clean_up(ca,bu,bl,x,cjac,c,clamda,objgrd,r,work,istate,iwork);
    error("memory allocation error");
  }

  /* Check the size of the linear constraint matrix */
  if (nclin > 0) {
    /* Get the user supplied matrix A, into CA */
    if ((failed = nag_extractDoubleMatrix(cmajor,nclin,n,a,ca))) {
       we04ucf_clean_up(ca,bu,bl,x,cjac,c,clamda,objgrd,r,work,istate,iwork);
      sprintf(err_msg,"'a' must be a %ld x %ld numeric matrix (code = %ld)", (long) nclin, (long) n, (long) failed);
      error(err_msg);
    }

    lda = nclin;
  } else {
    /* No linear constraints but needs to be allocated */
    lda = 1;
  }

  /* Allocate memory for the bounds */
  nbounds = n + nclin + ncnln;
  if (!(bu = (double *) malloc(nbounds*sizeof(double))) ||
      !(bl = (double *) malloc(nbounds*sizeof(double)))) {
     we04ucf_clean_up(ca,bu,bl,x,cjac,c,clamda,objgrd,r,work,istate,iwork);
    error("memory allocation error");
  }

  /* Copy and check the simple bounds */
  PROTECT(lower_s = VECTOR_ELT(lower,0));
  if ((failed = nag_extractDoubleVector(n,lower_s,bl))) {
     we04ucf_clean_up(ca,bu,bl,x,cjac,c,clamda,objgrd,r,work,istate,iwork);
    sprintf(err_msg,"First element of 'lower' must be a numeric vector of length %ld (code = %ld)", (long) n, (long) failed);
    error(err_msg);
  }

  PROTECT(upper_s = VECTOR_ELT(upper,0));
  if ((failed = nag_extractDoubleVector(n,upper_s,bu))) {
     we04ucf_clean_up(ca,bu,bl,x,cjac,c,clamda,objgrd,r,work,istate,iwork);
    sprintf(err_msg,"First element of 'upper' must be a numeric vector of length %ld (code = %ld)", (long) n, (long) failed);
    error(err_msg);
  }
  k = n;
  UNPROTECT(2);
  
  /* Copy and check the linear bounds (if required) */
  if (nclin > 0) {
    PROTECT(lower_l = VECTOR_ELT(lower,1));
    if ((failed = nag_extractDoubleVector(nclin,lower_l,&bl[k]))) {
       we04ucf_clean_up(ca,bu,bl,x,cjac,c,clamda,objgrd,r,work,istate,iwork);
      sprintf(err_msg,"Second element of 'lower' must be a numeric vector of length %ld (code = %ld)", (long) nclin, (long) failed);
      error(err_msg);
    }

    PROTECT(upper_l = VECTOR_ELT(upper,1));
    if ((failed = nag_extractDoubleVector(nclin,upper_l,&bu[k]))) {
       we04ucf_clean_up(ca,bu,bl,x,cjac,c,clamda,objgrd,r,work,istate,iwork);
      sprintf(err_msg,"Second element of 'upper' must be a numeric vector of length %ld (code = %ld)", (long) nclin, (long) failed);
      error(err_msg);
    }
    k += nclin;
    UNPROTECT(2);
  }

  /* Copy and check the non-linear bounds (if required) */
  if (ncnln > 0) {
    PROTECT(lower_n = VECTOR_ELT(lower,2));
    if ((failed = nag_extractDoubleVector(ncnln,lower_n,&bl[k]))) {
       we04ucf_clean_up(ca,bu,bl,x,cjac,c,clamda,objgrd,r,work,istate,iwork);
      sprintf(err_msg,"Third element of 'lower' must be a numeric vector of length %ld (code = %ld)", (long) ncnln, (long) failed);
      error(err_msg);
    }

    PROTECT(upper_n = VECTOR_ELT(upper,2));
    if ((failed = nag_extractDoubleVector(ncnln,upper_n,&bu[k]))) {
       we04ucf_clean_up(ca,bu,bl,x,cjac,c,clamda,objgrd,r,work,istate,iwork);
      sprintf(err_msg,"Third element of 'upper' must be a numeric vector of length %ld (code = %ld)", (long) ncnln, (long) failed);
      error(err_msg);
    }
    UNPROTECT(2);

    ldcj = ncnln;
    size_cjac = ldcj * n;
    size_c = ncnln;
  } else {
    /* There are no non-linear constraints */
    ldcj = 1;
    size_cjac = 1;
    size_c = 1;
  }

  /* Convert function pointer funct and rho into a couple of integers and store in IUSER */
  nag_convert_p2i(confun, &iuser[0]);
  nag_convert_p2i(objfun, &iuser[2]);
  nag_convert_p2i(rho, &iuser[4]);

  /* Allocate rest of memory */
  ldr = n;
  liwork = 3 * n + nclin + 2 * ncnln;
  lwork = 20 * n;
  if (ncnln > 0) {
    lwork = lwork + 2 * n * n + n * nclin + 2 * n * ncnln + 11 * nclin + 21 * ncnln;
  } else if (nclin > 0) {
    lwork = lwork + 2 * n * n + 11 * nclin;
  }
  if (!(x = (double *) malloc(n*sizeof(double))) ||
      !(cjac = (double *) malloc(size_cjac*sizeof(double))) ||
      !(c = (double *) malloc(size_c*sizeof(double))) ||
      !(istate = (NAG_fortran_integer *) malloc(nbounds*sizeof(NAG_fortran_integer))) ||
      !(clamda = (double *) malloc(nbounds*sizeof(double))) ||
      !(objgrd = (double *) malloc(nbounds*sizeof(double))) ||
      !(r = (double *) malloc(ldr*n*sizeof(double))) ||
      !(work = (double *) malloc(lwork*sizeof(double))) ||
      !(iwork = (NAG_fortran_integer *) malloc(liwork*sizeof(NAG_fortran_integer)))) {
     we04ucf_clean_up(ca,bu,bl,x,cjac,c,clamda,objgrd,r,work,istate,iwork);
    error("Memory allocation error");
  }    

  /* Get the initial values */
  if ((failed = nag_extractDoubleVector(n,initial,x))) {
     we04ucf_clean_up(ca,bu,bl,x,cjac,c,clamda,objgrd,r,work,istate,iwork);
    sprintf(err_msg,"'initial' must be a numeric vector of length %ld (code = %ld)", (long) n, (long) failed);
    error(err_msg);
  }

  /* Initialise the option arrays */
  ifail = 1;
  E04WBF(rname,6,cwsav,80,&lcwsav,lwsav,&llwsav,iwsav,&liwsav,rwsav,&lrwsav,&ifail);
  if (ifail != 0) {
    sprintf(err_msg,"E04WBF failed with IFAIL = %ld\n", (long) ifail);
     we04ucf_clean_up(ca,bu,bl,x,cjac,c,clamda,objgrd,r,work,istate,iwork);
    error(err_msg);
  }

  /* Handle options */
  unit_num = we04ucf_handle_options(optlist, lwsav, iwsav, rwsav);
  if (unit_num < 0) {
    /* Error opening output file */
    we04ucf_clean_up(ca,bu,bl,x,cjac,c,clamda,objgrd,r,work,istate,iwork);
    sprintf(err_msg,"Error opening monitoring file for output, check value for `monitoring file` entry in `optlist`");
    error(err_msg);
  } else if (unit_num > 0) {
    /* We are using an output file, so need to handle the advisory channel */

    /* Remember the current value */
    iflag = 0;
    X04ABF(&iflag, &nadv);

    /* Set the advisory channel to the output file. Need to do this, otherwise E04UCA sends output to both the
       advisory channel and the output file */
    iflag = 1;
    X04ABF(&iflag, &unit_num);
  }
    
  /* Call the NAG routine */
  ifail = 1;
  if (isLanguage(objfun)) {
    /* Run version that expects a body of a function */
    E04UCA(&n, &nclin, &ncnln, &lda, &ldcj, &ldr, ca, bl, bu, we04ucfL_con, we04ucfL_obj, &iter, istate, 
           c, cjac, clamda, &objf, objgrd, r, x, iwork, &liwork, work, &lwork, iuser, ruser, lwsav, 
           iwsav, rwsav, &ifail);

  } else { 
    /* Run version that expects the function itself */
    E04UCA(&n, &nclin, &ncnln, &lda, &ldcj, &ldr, ca, bl, bu, we04ucfF_con, we04ucfF_obj, &iter, istate, 
           c, cjac, clamda, &objf, objgrd, r, x, iwork, &liwork, work, &lwork, iuser, ruser, lwsav, 
           iwsav, rwsav, &ifail);
  }

  /* Error returns from NAG routine */
  if (ifail != 0) {
     we04ucf_clean_up(ca,bu,bl,x,cjac,c,clamda,objgrd,r,work,istate,iwork);
    sprintf(err_msg,"E04UCA failed with IFAIL = %ld\n", (long) ifail);
    error(err_msg);
  }

  /* Close the monitoring file if one opened */
  if (unit_num > 0) {
    ifail = 1;
    X04ADF(&unit_num, &ifail);

    /* Set the advisory channel back to its previous value */
    iflag = 1;
    X04ABF(&iflag, &nadv);
  }

  /* Set the number of values we are returning */
  lans = 7;
  if (ncnln > 0) {
    lans = lans + 2;
  }
  
  /* Allocate memory for output and the names of the output */
  PROTECT(ans = allocVector(VECSXP,lans));
  PROTECT(names = allocVector(STRSXP, lans));

  /* Get the list of results that are always returned */
  SET_VECTOR_ELT(ans,0,nag_d2r(1,&objf));
  SET_STRING_ELT(names,0,mkChar("objf"));
  SET_VECTOR_ELT(ans,1,nag_d2r(n,x)); 
  SET_STRING_ELT(names,1,mkChar("x")); 
  SET_VECTOR_ELT(ans,2,nag_d2r(n,objgrd)); 
  SET_STRING_ELT(names,2,mkChar("objgrd")); 
  SET_VECTOR_ELT(ans,3,nag_md2r(cmajor,n,n,r));
  SET_STRING_ELT(names,3,mkChar("r"));
  SET_VECTOR_ELT(ans,4,nag_fi2r(1,&iter));
  SET_STRING_ELT(names,4,mkChar("iter"));
  SET_VECTOR_ELT(ans,5,nag_fi2r(nbounds,istate));
  SET_STRING_ELT(names,5,mkChar("istate"));
  SET_VECTOR_ELT(ans,6,nag_d2r(nbounds,clamda));
  SET_STRING_ELT(names,6,mkChar("clamda"));
  if (ncnln > 0) {
    SET_VECTOR_ELT(ans,7,nag_d2r(ncnln,c));
    SET_STRING_ELT(names,7,mkChar("c"));

    SET_VECTOR_ELT(ans,8,nag_md2r(cmajor,ncnln,n,cjac));
    SET_STRING_ELT(names,8,mkChar("cjac"));
  }

  /* Set the names attribute */
  setAttrib(ans,R_NamesSymbol,names);

  /* Free up allocated memory */
  we04ucf_clean_up(ca,bu,bl,x,cjac,c,clamda,objgrd,r,work,istate,iwork);

  /* Unprotect the ans and names objects */
  UNPROTECT(2);

  return(ans);
}

/*********************************************************************************************/
/* Wrappers for callback functions, when user supplies the body of a function */
void __stdcall we04ucfL_obj(NAG_fortran_integer *mode, NAG_fortran_integer *n, double x[], double *objf, double objgrd[], 
                            NAG_fortran_integer *nstate, NAG_fortran_integer iuser[], double ruser[]) {
  /* we04ucfL_obj: Wrapper for the objective function callback function for E04UCF. For use
     when the user has supplied the body of a function (as opposed to the function itself).
     Used when calling a NAG Fortran routine

     On entry:
     mode   - Flag indicating what values the optimiser is expecting, 
               0 = obj, 1 = objgrd, 2 = obj and objgrd
     n      - The number of variables
     x      - The value at which the objective function and gradients should be evaluated, x[n]
     nstate - If NSTATE = 1 then the objective function is being called for the first time
     iuser  - Integer communication array
              iuser[0:1] - Holds integer representation of pointer to the function (or code segment)
                           that the user supplied for calculating the constraints
              iuser[2:3] - Holds integer representation of pointer to the function (or code segment)
                           that the user supplied for calculating objective function
              iuser[4:5] - Holds integer representation of pointer to the environment the functions
                           will be evaluated in
     ruser  - Real communication array, not used                           

     On exit:
     mode   - Set to -1000 if user function returned a list with an element called "terminate"
              Set to a different -ve value if the user function returned an invalid list, otherwise unchanged
     objf   - The objective function evaluated at x (if comm->flag = 0 or 2)
     objgrd - The gradients evaluated at x (if comm->flag = 2), objgrd[n]
  */
  SEXP objfun, rho, ans, tobj, tgrad, tstop;
  NAG_fortran_integer failed;

  /* Get the pointer to the R objective function */
  nag_convert_i2p(&iuser[2], &objfun);
  /* Get the pointer to the R environment */
  nag_convert_i2p(&iuser[4], &rho);

  /* Makes n, x, objgrd, flag and first available to the function as a variables of the same name
     NB: Need to make current values of conjac available, in case the user can only supply a few values for the
     derivatives
  */
  defineVar(install("n"), nag_fi2r(1,n), rho);
  defineVar(install("x"), nag_d2r(*n,x), rho);
  defineVar(install("objgrd"), nag_d2r(*n,objgrd), rho);
  defineVar(install("mode"), nag_fi2r(1,mode), rho);
  defineVar(install("nstate"), nag_fi2r(1,nstate), rho);

  /* Get the results from evaluating the function in R */
  PROTECT(ans = eval(objfun,rho));

  /* Check to see if user has included an element in the list called "terminate" */
  PROTECT(tstop = nag_getListElement(ans, "terminate"));
  if (tstop != R_NilValue) {
    /* User wants us to stop */
    *mode = -1000;

  } else {
    /* Expecting the result as a list, entries "objf" and "objgrd", holding the
       objective function and the gradients respectively.
       
       Could just use the first element in list as objf and second as objgrd, but
       this demonstrates use of nag_getListElement routine.
    */ 
    if (*mode == 0 || *mode == 2) {
      PROTECT(tobj = nag_getListElement(ans, "objf"));
      if (tobj != R_NilValue) {
        *objf = REAL(tobj)[0];
        if (!R_FINITE(*objf)) {
          /* Returned value is not valid, so kill the optimiser */
         *mode = -1;
        }
      }
      UNPROTECT(1);
    }

    if (*mode == 1 || *mode == 2) {
      PROTECT(tgrad = nag_getListElement(ans, "objgrd"));
      if (tgrad != R_NilValue) {
        if ((failed = nag_extractDoubleVector(*n,tgrad,objgrd))) {
          /* Returned value is not valid, so kill the optimiser */
          *mode = -(failed + 10);
        }
      }
      UNPROTECT(1);
    }
  }

  UNPROTECT(2);
}

void __stdcall we04ucfL_con(NAG_fortran_integer *mode, NAG_fortran_integer *ncnln, NAG_fortran_integer *n, NAG_fortran_integer *ldcj, 
                            NAG_fortran_integer needc[], double x[], double c[], double cjac[], NAG_fortran_integer *nstate, 
                            NAG_fortran_integer iuser[], double ruser[]) {
  /* we04ucfL_con: Wrapper for the constraint function callback function for E04UCF. For use
     when the user has supplied the body of a function (as opposed to the function itself).
     Used when calling a NAG Fortran routine

     On entry:
     mode   - Flag indicating what values the optimiser is expecting, 
               0 = c, 1 = cjac, 2 = c and cjac
     ncnln  -  The number of non-linear constraints
     n      - The number of variables
     ldcj   - Lead dimension of cjac (here we are making the assumption that ldcj = ncnln)
     needc  - If needc[i] > 0, then values for (i+1)'th constraint is required, needc[ncnln]
     x      - The value at which the constraints and gradients should be evaluated, x[n]
     iuser  - Integer communication array
              iuser[0:1] - Holds integer representation of pointer to the function (or code segment)
                           that the user supplied for calculating the constraints
              iuser[2:3] - Holds integer representation of pointer to the function (or code segment)
                           that the user supplied for calculating objective function
              iuser[4:5] - Holds integer representation of pointer to the environment the functions
                           will be evaluated in
     ruser  - Real communication array, not used                           

     On exit:
     mode   - Set to -1000 if user function returned a list with an element called "terminate"
              Set to a different -ve value if the user function returned an invalid list, otherwise unchanged
     c      - The constraints evaluated at x (if mode = 0 or 2), conf[ncnln]
     cjac   - The gradients evaluated at x (if mode = 1 or 2), cjac[ldcj*n], lcdj >= ncnln
  */
  SEXP confun, rho, ans, tconf, tcjac, tstop;
  NAG_fortran_integer failed;
  int cmajor;

  /* Get the pointer to the R constraint function */
  nag_convert_i2p(&iuser[0], &confun);
  /* Get the pointer to the R environment */
  nag_convert_i2p(&iuser[4], &rho);

  /* Makes n, ncnln, needc, x, cjac, mode and nstate available to the function as a variables of the same name.
     NB: Need to make current values of cjac available, in case the user can only supply a few values for the
     derivatives
  */
  cmajor = IS_COLUMN_MAJOR_ORDER;
  defineVar(install("n"), nag_fi2r(1,n), rho);
  defineVar(install("ncnln"), nag_fi2r(1,ncnln), rho);
  defineVar(install("needc"), nag_fi2r(*ncnln,needc), rho);
  defineVar(install("x"), nag_d2r(*n,x), rho);
  defineVar(install("cjac"), nag_md2r(cmajor,*ncnln,*n,cjac), rho);
  defineVar(install("mode"), nag_fi2r(1,mode), rho);
  defineVar(install("nstate"), nag_fi2r(1,nstate), rho);

  /* Get the results from evaluating the function in R */
  PROTECT(ans = eval(confun,rho));

  /* Check to see if user has included an element in the list called "terminate" */
  PROTECT(tstop = nag_getListElement(ans, "terminate"));
  if (tstop != R_NilValue) {
    /* User wants us to stop */
    *mode = -1000;

  } else {
    /* Expecting the result as a list, entries "c" and "cjac", holding the
       constraints and their derivatives respectively
    */
    if (*mode == 0 || *mode == 2) {
      PROTECT(tconf = nag_getListElement(ans, "c"));
      if (tconf != R_NilValue) {
        if ((failed = nag_extractDoubleVector(*ncnln,tconf,c))) {
          /* Returned value is not valid, so kill the optimiser */
          *mode = -(failed + 20);
        }
      }
      UNPROTECT(1);
    }
    
    if (*mode == 1 || *mode == 2) {
      PROTECT(tcjac = nag_getListElement(ans, "cjac"));
      if (tcjac != R_NilValue) {
        cmajor = IS_COLUMN_MAJOR_ORDER;
        if ((failed = nag_extractDoubleMatrix(cmajor,*ncnln,*n,tcjac,cjac))) {
          /* Returned value is not valid, so kill the optimiser */
          *mode = -(failed + 30);
        }
      }
      UNPROTECT(1);
    }
  }
  
  UNPROTECT(2);
}
/* End of wrappers for callback functions, when user supplies the body of a function */
/*********************************************************************************************/


/*********************************************************************************************/
/* Wrappers for callback functions, when user supplies the function itself */
void __stdcall we04ucfF_obj(NAG_fortran_integer *mode, NAG_fortran_integer *n, double x[], double *objf, double objgrd[], 
                            NAG_fortran_integer *nstate, NAG_fortran_integer iuser[], double ruser[]) {
  /* we04ucfF_obj: Wrapper for the objective function callback function for E04UCF. For use
     when the user has supplied the function itself (as opposed to the body of a function).
     Used when calling a NAG Fortran routine

     On entry:
     mode   - Flag indicating what values the optimiser is expecting, 
               0 = obj, 1 = objgrd, 2 = obj and objgrd
     n      - The number of variables
     x      - The value at which the objective function and gradients should be evaluated, x[n]
     nstate - If NSTATE = 1 then the objective function is being called for the first time
     iuser  - Integer communication array
              iuser[0:1] - Holds integer representation of pointer to the function (or code segment)
                           that the user supplied for calculating the constraints
              iuser[2:3] - Holds integer representation of pointer to the function (or code segment)
                           that the user supplied for calculating objective function
              iuser[4:5] - Holds integer representation of pointer to the environment the functions
                           will be evaluated in
     ruser  - Real communication array, not used                           

     On exit:
     mode   - Set to -1000 if user function returned a list with an element called "terminate"
              Set to a different -ve value if the user function returned an invalid list, otherwise unchanged
     objf   - The objective function evaluated at x (if comm->flag = 0 or 2)
     objgrd - The gradients evaluated at x (if comm->flag = 2), objgrd[n]
  */
  SEXP objfun, rho, ans, tobj, tgrad, tstop, R_fcall;
  NAG_fortran_integer failed;

  /* Get the pointer to the R objective function */
  nag_convert_i2p(&iuser[2], &objfun);
  /* Get the pointer to the R environment */
  nag_convert_i2p(&iuser[4], &rho);

  /* In this case objfun points to an actual function that takes 5 arguments */
  PROTECT(R_fcall = lang6(objfun,
                          nag_fi2r(1,n),
                          nag_d2r(*n,x),
                          nag_d2r(*n,objgrd),
                          nag_fi2r(1,mode),
                          nag_fi2r(1,nstate)));

  /* Get the results from evaluating the function in R */
  PROTECT(ans = eval(R_fcall,rho));

  /* Check to see if user has included an element in the list called "terminate" */
  PROTECT(tstop = nag_getListElement(ans, "terminate"));
  if (tstop != R_NilValue) {
    /* User wants us to stop */
    *mode = -1000;

  } else {
    /* Expecting the result as a list, entries "objf" and "objgrd", holding the
       objective function and the gradients respectively.
       
       Could just use the first element in list as objf and second as objgrd, but
       this demonstrates use of nag_getListElement routine.
    */ 
    if (*mode == 0 || *mode == 2) {
      PROTECT(tobj = nag_getListElement(ans, "objf"));
      if (tobj != R_NilValue) {
        *objf = REAL(tobj)[0];
        if (!R_FINITE(*objf)) {
          /* Returned value is not valid, so kill the optimiser */
         *mode = -1;
        }
      }
      UNPROTECT(1);
    }

    if (*mode == 1 || *mode == 2) {
      PROTECT(tgrad = nag_getListElement(ans, "objgrd"));
      if (tgrad != R_NilValue) {
        if ((failed = nag_extractDoubleVector(*n,tgrad,objgrd))) {
          /* Returned value is not valid, so kill the optimiser */
          *mode = -(failed + 10);
        }
      }
      UNPROTECT(1);
    }
  }

  UNPROTECT(3);
}

void __stdcall we04ucfF_con(NAG_fortran_integer *mode, NAG_fortran_integer *ncnln, NAG_fortran_integer *n, NAG_fortran_integer *ldcj, 
                            NAG_fortran_integer needc[], double x[], double c[], double cjac[], NAG_fortran_integer *nstate, 
                            NAG_fortran_integer iuser[], double ruser[]) {
  /* we04ucfF_con: Wrapper for the constraint function callback function for E04UCF. For use
     when the user has supplied the function itself (as opposed to the body of a function).
     Used when calling a NAG Fortran routine

     On entry:
     mode   - Flag indicating what values the optimiser is expecting, 
               0 = c, 1 = cjac, 2 = c and cjac
     ncnln  -  The number of non-linear constraints
     n      - The number of variables
     ldcj   - Lead dimension of cjac (here we are making the assumption that ldcj = ncnln)
     needc  - If needc[i] > 0, then values for (i+1)'th constraint is required, needc[ncnln]
     x      - The value at which the constraints and gradients should be evaluated, x[n]
     iuser  - Integer communication array
              iuser[0:1] - Holds integer representation of pointer to the function (or code segment)
                           that the user supplied for calculating the constraints
              iuser[2:3] - Holds integer representation of pointer to the function (or code segment)
                           that the user supplied for calculating objective function
              iuser[4:5] - Holds integer representation of pointer to the environment the functions
                           will be evaluated in
     ruser  - Real communication array, not used                           

     On exit:
     mode   - Set to -1000 if user function returned a list with an element called "terminate"
              Set to a different -ve value if the user function returned an invalid list, otherwise unchanged
     c      - The constraints evaluated at x (if mode = 0 or 2), conf[ncnln]
     cjac   - The gradients evaluated at x (if mode = 1 or 2), cjac[ldcj*n], lcdj >= ncnln
  */
  SEXP confun, rho, ans, tconf, tcjac, tstop, R_fcall;
  NAG_fortran_integer failed;
  int cmajor;

  /* Get the pointer to the R constraint function */
  nag_convert_i2p(&iuser[0], &confun);
  /* Get the pointer to the R environment */
  nag_convert_i2p(&iuser[4], &rho);

  /* Makes n, ncnln, needc, x, cjac, mode and nstate available to the function as a variables of the same name.
     NB: Need to make current values of cjac available, in case the user can only supply a few values for the
     derivatives
  */
  cmajor = IS_COLUMN_MAJOR_ORDER;
  PROTECT(R_fcall = lang8(confun,
                          nag_fi2r(1,n),
                          nag_fi2r(1,ncnln),
                          nag_fi2r(*ncnln,needc),
                          nag_d2r(*n,x),
                          nag_md2r(cmajor,*ncnln,*n,cjac),
                          nag_fi2r(1,mode),
                          nag_fi2r(1,nstate)));

  /* Get the results from evaluating the function in R */
  PROTECT(ans = eval(R_fcall,rho));

  /* Check to see if user has included an element in the list called "terminate" */
  PROTECT(tstop = nag_getListElement(ans, "terminate"));
  if (tstop != R_NilValue) {
    /* User wants us to stop */
    *mode = -1000;

  } else {
    /* Expecting the result as a list, entries "c" and "cjac", holding the
       constraints and their derivatives respectively
    */
    if (*mode == 0 || *mode == 2) {
      PROTECT(tconf = nag_getListElement(ans, "c"));
      if (tconf != R_NilValue) {
        if ((failed = nag_extractDoubleVector(*ncnln,tconf,c))) {
          /* Returned value is not valid, so kill the optimiser */
          *mode = -(failed + 20);
        }
      }
      UNPROTECT(1);
    }
    
    if (*mode == 1 || *mode == 2) {
      PROTECT(tcjac = nag_getListElement(ans, "cjac"));
      if (tcjac != R_NilValue) {
        cmajor = IS_COLUMN_MAJOR_ORDER;
        if ((failed = nag_extractDoubleMatrix(cmajor,*ncnln,*n,tcjac,cjac))) {
          /* Returned value is not valid, so kill the optimiser */
          *mode = -(failed + 30);
        }
      }
      UNPROTECT(1);
    }
  }
  
  UNPROTECT(3);
}
/* End of wrappers for callback functions, when user supplies the function itself */
/*********************************************************************************************/

int we04ucf_handle_options(SEXP optlist, NAG_fortran_integer lwsav[], NAG_fortran_integer iwsav[], double rwsav[]) {
  /* we04ucf_handle_options: Take an R list, and extract from it any options for E04UCF.

    On entry:
    optlist  - An R list
    options  - Pointer to pointer for a variable of type Nag_E04_Opt

    On exit:
    options  - If optlist is a list, then the elements of options are set
               based on the names of the values, and the values themselves, in the list.
               If an output file has been specified, then routine returns the unit number
               associated with the file (allowing us to close it). A value of 0 returned
               means that no file was specified. A -ve value means there was an error 
               opening the file.

               So, optlist = list(list=TRUE,iters=30), would set the option "list"  
               the option "iters" to 30. Only the first entry of each entry in optlist is used. 
               The items in the list are processed in order. The names are the same as listed in the e04ucf doc
               (all lower case). 

               The exception to the above is the "monitoring file" option. Rather than specify a unit
               number (as described in the doc), a file name should be supplied, i.e.
               list("monitoring file" = "tt.txt") sends output to tt.txt

    Returns:
    0 if no file opened, -1 if an error occurred opening the file and the unit number if a file
    has been opened
  */
  SEXP names, tentry;
  int i, ival;
  NAG_fortran_integer ifail, slen;
  double dval;
  const char *this_name;
  char str[100];
  int unit_num = 0, mode = 1;

  /* No options specified */
  if (!isNewList(optlist)) {
    return 0;
  }

  /* Get the list of names from the list */
  PROTECT(names = getAttrib(optlist,R_NamesSymbol));
  for (i = 0; i < length(optlist); i++) {
    this_name = CHAR(STRING_ELT(names,i));
        
    /* Only interested in the first element, which is what asReal should
       convert to a double */
    PROTECT(tentry = VECTOR_ELT(optlist,i));
    if (!isString(tentry)) {
      dval = asReal(tentry);
      if (!R_FINITE(dval)) dval = 0.0;
      ival = (int) dval;
    } else {
      dval = -1.0;
      ival = -1;
    }

    str[0] = '\0';    
    if (!strcmp(this_name,"list")) {
      if (ival) {
        sprintf(str,"list");
      } else {
        sprintf(str,"nolist");
      }
    } else if (!strcmp(this_name,"unit number")) {
      /* This isn't an e04ucf option, but allows us to change the default unit number for
         monitoring files if required */
      unit_num = ival;
    } else if (!strcmp(this_name,"file mode")) {
      /* This isn't an e04ucf option, but allows use to append to a specified monitoring
         file by specifying a value of 2. All other values cause a new file to be created */
      mode = ival;      
    } else if (!strcmp(this_name,"central difference interval") || !strcmp(this_name,"ce")) {
      sprintf(str,"central difference interval = %g", dval);
    } else if (!strcmp(this_name, "crash tolerance")) {
      sprintf(str,"crash_tolerance = %g", dval);
    } else if (!strcmp(this_name, "defaults")) {
      if (ival) {
        sprintf(str,"defaults");
      }
    } else if (!strcmp(this_name, "derivative level") || !strcmp(this_name, "der")) {
      sprintf(str,"derivative level = %d" , ival);
    } else if (!strcmp(this_name, "difference interval") || !strcmp(this_name, "diff")) {
      sprintf(str,"difference interval = %g", dval);
    } else if (!strcmp(this_name, "feasibility tolerance") ||!strcmp(this_name, "fe")) {
      sprintf(str,"feasibility tolerance = %g", dval);
    } else if (!strcmp(this_name, "function precision") || !strcmp(this_name, "fu")) {
      sprintf(str,"function precision = %g", dval);
    } else if (!strcmp(this_name, "hessian") || !strcmp(this_name, "h")) {
      sprintf(str,"hessian = %s", (ival) ? "Yes" : "No");
    } else if (!strcmp(this_name, "infinite bound size") || !strcmp(this_name, "inb")) {
      sprintf(str,"infinite bound size = %g", dval);
    } else if (!strcmp(this_name, "infinite step size") || !strcmp(this_name, "ins")) {
      sprintf(str,"infinite step size = %g", dval);
    } else if (!strcmp(this_name, "iters") || !strcmp(this_name, "iter")) {
      sprintf(str,"iters = %d", ival);
    } else if (!strcmp(this_name, "iteration limit") || !strcmp(this_name, "itera")) {
      sprintf(str,"iteration limit = %d", ival);
    } else if (!strcmp(this_name, "itns")) {
      sprintf(str,"itns = %d", ival);
    } else if (!strcmp(this_name, "line search tolerance") || !strcmp(this_name, "line")) {
      sprintf(str,"line search tolerance = %g", dval);
    } else if (!strcmp(this_name, "linear feasibility tolerance") || !strcmp(this_name, "linearf")) {
      sprintf(str,"linear feasibility tolerance = %g", dval);
    } else if (!strcmp(this_name, "major iteration limit") || !strcmp(this_name, "mait")) {
      sprintf(str,"major iteration limit = %d", ival);
    } else if (!strcmp(this_name, "major print level") || !strcmp(this_name, "majp")) {
      sprintf(str,"major print level = %d", ival);
    } else if (!strcmp(this_name, "minor iteration limit") || !strcmp(this_name, "minitera")) {
      sprintf(str,"minor iteration limit = %d", ival);
    } else if (!strcmp(this_name, "minor print level") || !strcmp(this_name, "minp")) {
      sprintf(str,"minor print level = %d", ival);
    } else if (!strcmp(this_name, "monitoring file") || !strcmp(this_name, "mo")) {      
      /* Can't have mode 0 as this is reading mode only */
      if (mode == 0) mode = 1;
      /* Use unit_num 7 by default */
      if (unit_num == 0) unit_num = 7;
      /* Open up a file for output (either write or append) */
      ifail = 1;
      
      X04ACF(&unit_num,strncpy(str,CHAR(STRING_ELT(tentry,0)),100),100, &mode,&ifail);
      if (ifail != 0) {
        unit_num = -1;       
        break;
      }
      sprintf(str,"monitoring file = %d", unit_num);
    } else if (!strcmp(this_name, "nonlinear feasibility tolerance") || !strcmp(this_name, "nonf")) {
      sprintf(str,"nonlinear feasibility tolerance = %g", dval);
    } else if (!strcmp(this_name, "optimality tolerance") || !strcmp(this_name, "op")) {
      sprintf(str,"optimality tolerance = %g", dval);
    } else if (!strcmp(this_name, "print level") || !strcmp(this_name, "print")) {
      sprintf(str,"print level = %d", ival);
    } else if (!strcmp(this_name, "step limit") || !strcmp(this_name, "st")) {
      sprintf(str,"step limit = %g", dval);
    } else if (!strcmp(this_name, "start constraint check at variable") || !strcmp(this_name, "stac")) {
      sprintf(str,"start constraint check at variable = %d", ival);
    } else if (!strcmp(this_name, "start objective check at variable") || !strcmp(this_name, "stao")) {
      sprintf(str,"start objective check at variable = %d", ival);
    } else if (!strcmp(this_name, "stop constraint check at variable") || !strcmp(this_name, "stoc")) {
      sprintf(str,"stop constraint check at variable = %d",ival);
    } else if (!strcmp(this_name, "stop objective check at variable") || !strcmp(this_name, "stoo")) {
      sprintf(str,"stop objective check at variable = %d", ival);
    } else if (!strcmp(this_name, "verify") || !strcmp(this_name, "ve")) {
      sprintf(str,"verify = %d", ival);
    } else if (!strcmp(this_name, "verify constraints gradient") || !strcmp(this_name, "vec")) {
      sprintf(str,"verify constraints gradient = %d", ival);
    } else if (!strcmp(this_name, "verify gradient") || !strcmp(this_name, "veg")) {
      sprintf(str,"verify gradient = %d", ival);
    } else if (!strcmp(this_name, "verify level") || !strcmp(this_name, "vel")) {
      sprintf(str,"verify level = %d", ival);
    } else if (!strcmp(this_name, "verify objective gradient") || !strcmp(this_name, "veo")) {
      sprintf(str,"verify objective gradient = %d", ival);
    }

    /* All other options are ignored (this includes the valid ones: cold start and warm start) */

    UNPROTECT(1);

    /* Call the option setting routine */
    if (str[0]) {
      ifail = 1;
      slen = strlen(str);
      E04UEA(str,slen,lwsav,iwsav,rwsav,&ifail);
    }
  }
  
  UNPROTECT(1);

  return unit_num;
}


void we04ucf_clean_up(double *pd1, double *pd2, double *pd3, double *pd4, double *pd5, double *pd6, double *pd7,
                      double *pd8, double *pd9, double *pd10, NAG_fortran_integer *pi1, NAG_fortran_integer *pi2) {
  /* we04ucf_clean_up: Frees any allocated memory.

     Routine frees the various pointers, if they are non-null
  */

  if (pd1) free(pd1);
  if (pd2) free(pd2);
  if (pd3) free(pd3);
  if (pd4) free(pd4);
  if (pd5) free(pd5);
  if (pd6) free(pd6);
  if (pd7) free(pd7);
  if (pd8) free(pd8);
  if (pd9) free(pd9);
  if (pd10) free(pd10);
  if (pi1) free(pi1);
  if (pi2) free(pi2);
  return;
}
